#!/usr/bin/python3
__author__ = 'Przemek Decewicz'

import gzip
import sys
import pkg_resources
from argparse import ArgumentParser, RawDescriptionHelpFormatter
from Bio import SeqIO
from compare_predictions_to_phages import genbank_seqio
from os import makedirs, path
from PhiSpyModules import log_and_message, is_gzip_file

from argparse import ArgumentParser
from os import makedirs, path
from sys import argv

def read_prophage_coordinates(infile):
    """
    Reads PhiSpy's prophage coordinates file.
    :param infile: path to the file
    :return prophages: dictionary of replicons ids and prophage coordinates
    """

    prophages = {}

    with open(infile) as inf:
        for line in inf:
            pp, replicon_id, start, end, r1_start, r1_end, r2_start, r2_end, r1_seq, r2_seq, note = line.strip().split('\t')

            # adjust coords to Python's indexing
            coords = (int(start) - 1, int(end) - 1)

            try:
                prophages[replicon_id][coords] = pp
            except KeyError:
                prophages[replicon_id] = {coords: pp}

    return prophages

def read_prophage_table(infile):
    """
    Reads tab-delimited table with columns for:
    1 - path to GenBank file,
    2 - replicon id,
    3 - prophage start coordinate,
    4 - prophage end coordinate,
    5 (optional) - prophage name (if not provided pp1, pp2, etc.
    will be assigned for each file)
    :param infile: path to the file
    :return prophages: dictionary of GenBank file(s) paths, replicons and prophages coordinates
    """

    prophages = {}
    with open(infile) as inf:
        for line in inf:
            if line.startswith('#'): continue

            line = line.strip().split('\t')
            file_path, replicon_id, start, end = line[:4]

            # adjust coords to Python's indexing
            coords = (int(start) - 1, int(end) - 1)

            if start > end:
                start, end = end, start

            if file_path in prophages:
                pp_cnt = len(prophages[file_path]) + 1
            else:
                pp_cnt = 1

            pp = line[4] if len(line) == 5 else  f"pp{pp_cnt}"

            try:
                prophages[file_path][replicon_id][coords] = pp
            except KeyError:
                try:
                    prophages[file_path][replicon_id] = {coords: pp}
                except KeyError:
                    prophages[file_path] = {replicon_id: {coords: pp}}

    return prophages

def feature_within_prophage(feature, pps):
    """
    Checks whether a feature is located within any of prophage regions.
    :param feature: SeqFeature object
    :param pps: dictionary of prophage coordinates tuples
    :return pp: prophage name
    """

    f_start = int(feature.location.start)
    f_end = int(feature.location.end)

    for pp_coords in pps:
        if f_start >= pp_coords[0] and f_end <= pp_coords[1]:
            return pps[pp_coords]

    return None


def mark_prophage_features(prophages, outdir, gzip_files, ungzip_files):
    """
    Marks features within prophage regions with "is_phage='1'" qualifier.
    :param prophages: dictionary of GenBank file(s) paths, replicons and prophages coordinates
    :param outdir: output directory to write updated GenBank file(s)
    :param gzip_files: gzip all processed input files
    :param ungzip_files: ungzip all processed input files
    :return:
    """

    for gbkf, pps in prophages.items():
        log_and_message(f"Updating {path.basename(gbkf)}.", c="YELLOW", stderr = True)
        records = list(genbank_seqio(gbkf))
        cnt_all = 0
        cnt_cds = 0
        for record in records:
            if record.id in pps:
                for feature in record.features:
                    pp = feature_within_prophage(feature, pps[record.id])
                    if pp:
                        feature.qualifiers['is_phage'] = ['1']
                        feature.qualifiers['color'] = ['6']
                        try:
                            feature.qualifiers['note'].append(pp)
                        except KeyError:
                            feature.qualifiers['note'] = [pp]

                        cnt_all += 1
                        if feature.type == 'CDS': cnt_cds += 1

        log_and_message(f"-- Marked {cnt_all} features, including {cnt_cds} CDSs, with 'is_phage' qualifier.", stderr=True)

        outfile = path.join(outdir, path.basename(gbkf))
        if gzip_files:
            if not outfile.endswith('.gz'):
                outfile += '.gz'
            handle = gzip.open(outfile, 'wt')
        elif ungzip_files:
            if outfile.endswith('.gz'):
                outfile = outfile[:-3]
            handle = open(outfile, 'w')
        elif is_gzip_file(gbkf):
            handle = gzip.open(outfile, 'wt')
        else:
            handle = open(outfile, 'w')

        log_and_message(f"Writing {outfile}", c="GREEN", stderr=True)
        SeqIO.write(records, handle, 'genbank')


def main():
    args = ArgumentParser(prog = "mark_prophage_features.py",
                          description = "Updates GenBank files based on prophage_coordinates.tsv file or other tab-delimited table and marks features within indicated coordinates with 'is_phage=1' qualifier.",
                          epilog = "Example usages:\npython3 scripts/mark_prophage_features.py --genbank infile.gbk --outdir updated_genbanks --ppcoords PhiSpy_output/prophage_coordinates.tsv\npython3 scripts/mark_prophage_features.py --outdir updated_genbanks --table my_confirmed_predictions.tsv",
                          formatter_class = RawDescriptionHelpFormatter)

    args.add_argument('-g', '--genbank',
                      type = str,
                      help = 'Path to input GenBank file.')

    args.add_argument('-o', '--outdir',
                      type = str,
                      help = 'Path to output directory to write updated GenBank(s).',
                      required = True)

    args.add_argument('-c', '--ppcoords',
                      type = str,
                      help = 'Path to prophage_coordinates.tsv file.')

    args.add_argument('-t', '--table',
                      type = str,
                      help = 'Path to tab-delimited file with confirmed prophage regions to mark. The file has to have the following columns: 1 - path to GenBank file, 2 - replicon id, 3 - prophage start coordinate, 4 - prophage end coordinate, 5 (optional) - prophage name (if not provided pp1, pp2, etc. will be assigned for each file)')

    args.add_argument('--gzip_files',
                      action = 'store_true',
                      default = False,
                      help = 'Gzip all output files. \'.gz\' extension will be added if missing. [Default: %(default)s]')

    args.add_argument('--ungzip_files',
                      action = 'store_true',
                      default = False,
                      help = 'Ungzip all output files. \'.gz\' extension will be removed if present. [Default: %(default)s]')

    if len(argv[1:]) == 0:
        args.print_help()
        args.exit()

    try:
        args = args.parse_args()
    except:
        args.exit()

    if not path.isdir(args.outdir): makedirs(args.outdir)

    if args.genbank and args.outdir and args.ppcoords:
        log_and_message(f"Marking prophage features for a single GenBank file based on PhiSpy's prophage coordinates file.", c="GREEN", stderr=True)

        prophages = {args.genbank: read_prophage_coordinates(args.ppcoords)}

    elif args.outdir and args.table:
        log_and_message(f"Marking prophage features for GenBank file(s) based on prophage table.", c="GREEN", stderr=True)

        prophages = read_prophage_table(args.table)

    else:
        log_and_message(f"Incorrect input data.", c="RED", stderr=True)
        log_and_message(f"Use --genbank, --outdir and --ppcords or --table and --outdir.", c="PINK", stderr=True)

    mark_prophage_features(prophages, args.outdir, args.gzip_files, args.ungzip_files)

    log_and_message(f"Done. You can now use output file(s) for training PhiSpy.", c="GREEN", stderr=True)

if __name__ == '__main__':
    main()



"""
An alternative search phmms method that does not require a separate reading of the file. We parse
the data record we already have.
Need to adjust the location of the phmms call in PhiSpy.py
"""

import os
import sys

from io import StringIO
import subprocess
from argparse import Namespace
from tempfile import NamedTemporaryFile
from .genbank_accessory_functions import feature_id

from .log_and_message import log_and_message

from Bio import SearchIO

def search_phmms(**kwargs):
    """
    This is an alternate implementation of the search phmms
    but we write to a temp file and then parse that.
    See hmms/run_hmmer.py in EdwardsLab git repo for timing tests
    of different ways of doing this!
    :param kwargs:
    :return:
    """

    self = Namespace(**kwargs)
    log_and_message(f"Running HMM profiles against {self.phmms}", c="GREEN", stderr=True, quiet=self.quiet)
    # write the amino acids to a named temporary file that we will unlink later
    aaout = NamedTemporaryFile(mode='w+t', dir=self.output_dir, delete=False)
    log_and_message(f"hmmsearch: writing the amino acids to temporary file {aaout.name}\n", c="GREEN", stderr=True, quiet=self.quiet)
    aaout.seek(0)
    all_features = {}
    for seq in self.record:
        for feat in seq.get_features('CDS'):
            aa = ""
            if 'translation' in feat.qualifiers:
                aa = feat.qualifiers['translation'][0]
            else:
                aa = str(feat.extract(seq).translate().seq)
                feat.qualifiers['translation'] = [aa]
            myid = feature_id(seq, feat)
            if myid in all_features:
                log_and_message(f"WARNING: {myid} is not a unique id. We need unique protein IDs to run hmmsearch.\nThis feature will be skipped during the search.\n",
                    c="RED", stderr=True, quiet=False)
                continue
            aaout.write(f">{myid}\n{aa}\n")
            all_features[myid] = feat
    aaout.close()

    if len(all_features) == 0:
        log_and_message(f"FATAL: none of protein features had protein ID - can't run hmmsearch.", 
            c="RED", stderr=True, quiet=False)
        sys.exit(-1)
    else:
        log_and_message(f"Searching {len(all_features)} proteins with hmmsearch.", c="GREEN", stderr=True, quiet=False)

    try:
        search = subprocess.Popen(["hmmsearch", '--cpu', str(self.threads), '-E', '1e-10', '--domE', '1e-5', '--noali', self.phmms, aaout.name], stdout=subprocess.PIPE)
    except subprocess.CalledProcessError as e:
        sys.stderr.write(f"Error running hmmscan:\n{e}\n")
        sys.exit(-1)

    hmmresult = search.communicate()[0]
    results = SearchIO.parse(StringIO(hmmresult.decode()), 'hmmer3-text')

    for res in results:
        for hit in res:
            if hit.id not in all_features:
                log_and_message(f"Found an HMM hit to {hit.id} but we don't have a record of that feature", c="RED", stderr=True,
                                quiet=False)
                continue
            if 'phmm' not in all_features[hit.id].qualifiers:
                all_features[hit.id].qualifiers['phmm'] = []
            all_features[hit.id].qualifiers['phmm'].append(f"{res.id}:{hit.evalue}")
            # color the gene, unless it is already colored. We overwrite a hypothetical colour
            if self.color and 'colour' not in all_features[hit.id].qualifiers or all_features[hit.id].qualifiers['colour'] == 13:
                all_features[hit.id].qualifiers['colour'] = 9 # light blue

    # remove the amino acids
    os.unlink(os.path.join(self.output_dir, aaout.name))
    log_and_message(f"Completed running HMM profiles against {self.phmms}", c="GREEN", stderr=True, quiet=self.quiet)
    return self.record



"""
Compare the regions predicted to be prophages to the regions that are marked as prophages in our testing set

Probably the hardest part of this is the identifiers!
"""

import os
import sys
import argparse
import gzip
from Bio import SeqIO, BiopythonWarning
from PhiSpyModules import message, is_gzip_file

__author__ = 'Rob Edwards'
__copyright__ = 'Copyright 2020, Rob Edwards'
__credits__ = ['Rob Edwards']
__license__ = 'MIT'
__maintainer__ = 'Rob Edwards'
__email__ = 'raedwards@gmail.com'


def genbank_seqio(gbkf):
    if is_gzip_file(gbkf):
        handle = gzip.open(gbkf, 'rt')
    else:
        handle = open(gbkf, 'r')

    return SeqIO.parse(handle, "genbank")

def actual_phage_cds(gbkf, verbose=False):
    """
    Read the genbank file and return a list of features that are actually phage regions
    :param gbkf: the test genbank file with CDS marked with is_phage
    :param verbose: more output
    :return: a set of phage features
    """

    if verbose:
        message(f"Reading {gbkf}", "GREEN", "stderr")

    phage = {}
    nonphage = {}
    for seq in genbank_seqio(gbkf):
        for feat in seq.features:
            if feat.type == 'CDS':
                if 'product' not in feat.qualifiers:
                    feat.qualifiers['product'] = [f"Hypothetical protein (not annotated in {gbkf})"]
                if 'is_phage' in feat.qualifiers:
                    phage[str(feat.translate(seq, cds=False).seq).upper()] = feat.qualifiers['product'][0]
                else:
                    nonphage[str(feat.translate(seq, cds=False).seq).upper()] = feat.qualifiers['product'][0]

    return phage, nonphage

def predicted_genbank(predf, verbose=False):
    """
    Read the predictions from the genbank file and return
    a set of features
    :param predf: the predictions file
    :param verbose: more output
    :return: a set of predicted phage genes
    """

    if verbose:
        message(f"Reading {predf}", "GREEN", "stderr")

    predicted = {}
    for seq in genbank_seqio(predf):
        for feat in seq.features:
            if feat.type == 'CDS':
                if 'product' in feat.qualifiers:
                    predicted[str(feat.translate(seq, cds=False).seq).upper()] = feat.qualifiers['product'][0]
                else:
                    predicted[str(feat.translate(seq, cds=False).seq).upper()] = f"Hypothetical protein (not annotated in {predf})"

    if verbose:
        message(f"Found {len(predicted)} predicted prophage features", "BLUE", "stderr")

    return predicted

def predicted_regions(regf, gbkf, verbose):
    """
    Pull the phage genes from the regions
    :param regf: the regions file with contigs/start/stop
    :param gbkf: the genbank file used to make those predictions
    :param verbose: more output
    :return: a set of predicted phage genes
    """

    regions = {}
    if verbose:
        message(f"Reading {regf}", "GREEN", "stderr")
    with open(regf, 'r') as f:
        for l in f:
            p = l.strip().split("\t")
            assert(len(p) == 3), f"Expected a tple of [contig, start, stop] in {regf}"
            p[1] = int(p[1])
            p[2] = int(p[2])
            if p[0] not in regions:
                regions[p[0]] = []
            if p[2] < p[1]:
                regions[p[0]].append([p[2], p[1]])
            else:
                regions[p[0]].append([p[1], p[2]])

    if verbose:
        message(f"Reading {gbkf} again to get the phage regions", "GREEN", "stderr")

    predicted = {}
    for seq in genbank_seqio(gbkf):
        if seq.id in regions:
            for loc in regions[seq.id]:
                if verbose:
                    message(f"Getting from {loc[0]} to {loc[1]}", "PINK", "stderr")
                for feat in seq[loc[0]:loc[1]].features:
                    if feat.type == 'CDS':
                        if 'product' in feat.qualifiers:
                            predicted[str(feat.translate(seq[loc[0]:loc[1]], cds=False).seq).upper()] = feat.qualifiers['product'][0]
                        else:
                            predicted[str(feat.translate(seq[loc[0]:loc[1]], cds=False).seq).upper()] = f"Hypothetical protein (not annotated in {gbkf})"

    if verbose:
        message(f"Found {len(predicted)} predicted prophage features", "BLUE", "stderr")

    return predicted

def compare_real_predicted(phage: dict, nonphage: dict, predicted: dict, print_fp: bool, print_fn: bool, verbose: bool):
    """
    Compare the features that are real and predicted
    :param print_fn: print out the false negative matches
    :param print_fp: print out the false positive matches
    :param phage: actual phage features
    :param nonphage: actual non phage features
    :param predicted: predicted phage features
    :param verbose: more output
    :return:
    """

    if verbose:
        message(f"Comparing real and predicted", "GREEN", "stderr")

    # TP = phage intersection predicted
    # TN = nonphage intersection [not in predicted]
    # FP = nonphage intersection predicted
    # FN = phage intersection [not in predicted]

    # convert the keys to sets
    phage_set = set(phage.keys())
    nonphage_set = set(nonphage.keys())
    predicted_set = set(predicted.keys())

    # calculate not in predicted
    not_predicted = set()
    for s in phage_set.union(nonphage):
        if s not in predicted:
            not_predicted.add(s)

    print(f"Found:\nTest set:\n\tPhage: {len(phage)} Not phage: {len(nonphage)}")
    print(f"Predictions:\n\tPhage: {len(predicted)} Not phage: {len(not_predicted)}")

    tp = len(phage_set.intersection(predicted_set))
    tn = len(nonphage_set.intersection(not_predicted))
    fp = len(nonphage_set.intersection(predicted_set))
    fn = len(phage_set.intersection(not_predicted))

    print(f"TP: {tp}  FP: {fp}  TN: {tn}  FN: {fn}")
    try:
        accuracy = (tp+tn)/(tp + tn + fp + fn)
    except ZeroDivisionError:
        accuracy = "NaN"

    try:
        precision = tp/(tp+fp)
    except ZeroDivisionError:
        precision = "NaN"

    try:
        recall = tp/(tp+fn)
    except ZeroDivisionError:
        recall = "NaN"

    try:
        specificity = tn/(tn+fp)
    except ZeroDivisionError:
        specificity = "NaN"

    f1_score = "NaN"
    if accuracy != "NaN" and precision != "NaN" and recall != "NaN" and specificity != "NaN":
        try:
            f1_score = 2*(recall * precision) / (recall + precision)
        except ZeroDivisionError:
            f1_score = "NaN"

    if accuracy != "NaN":
        print(f"Accuracy:    {accuracy:.3f}\t(this is the ratio of the correctly labeled phage genes to the whole pool of genes")
    else:
        print("Accuracy: NaN")

    if precision != "NaN":
        print(f"Precision:   {precision:.3f}\t(This is the ratio of correctly labeled phage genes to all predictions)")
    else:
        print("Precision: NaN")

    if recall != "NaN":
        print(f"Recall:      {recall:.3f}\t(This is the fraction of actual phage genes we got right)")
    else:
        print("Recall: NaN")

    if specificity != "NaN":
        print(f"Specificity: {specificity:.3f}\t(This is the fraction of non phage genes we got right)")
    else:
        print("Specificity: NaN")
    
    if f1_score != "NaN":
        print(f"f1 score: {f1_score:.3f}\t(this is the harmonic mean of precision and recall, and is the best measure when, as in this case, there is a big difference between the number of phage and non-phage genes)")
    else:
        print("f1 score: NaN")

    if print_fp:
        for i in nonphage_set.intersection(predicted_set):
            print(f"FP\t{i}\t{nonphage[i]}")

    if print_fn:
        for i in phage_set.intersection(not_predicted):
            print(f"FN\t{i}\t{[phage[i]]}")



if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Compare predictions to reality")
    parser.add_argument('-t', '--testfile', help='test file that has phage proteins marked with the is_phage qualifier', required=True)
    parser.add_argument('-p', '--predictfile', help='predictions genbank file that has each prophage as a sequence entry')
    parser.add_argument('-r', '--regionsfile', help='predictions regions file that has tuple of [contig, start, end]')
    parser.add_argument('--fp', help='print out the false positives', action='store_true')
    parser.add_argument('--fn', help='print out the false negatives', action='store_true')
    parser.add_argument('-v', help='verbose output', action='store_true')
    args = parser.parse_args()

    pred = None
    if args.predictfile:
        pred = predicted_genbank(args.predictfile, args.v)
    elif args.regionsfile:
        pred = predicted_regions(args.regionsfile, args.testfile, args.v)
    else:
        message("FATAL: Please provide either a predictions genbank or tsv file", "RED", "stderr")

    phage, nonphage = actual_phage_cds(args.testfile, args.v)

    compare_real_predicted(phage, nonphage, pred, args.fp, args.fn, args.v)
